/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Prevents multiple definitions
#ifndef WRSTATION_COMMANDSERIALINTERFACE_COMMANDINTERFACE
#define WRSTATION_COMMANDSERIALINTERFACE_COMMANDINTERFACE


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>
#include <QtSerialPort/QtSerialPort>


// TODO: Insert class and method descriptions
class WRStation::CommandSerialInterface::CommandInterface : public QObject {
    Q_OBJECT

    private:
        QHash< quint8, QHash<quint8, bool> >  _controlledPlayers;
        QHash< quint8, QHash<quint8, CommandPlayerState> > _controlledPlayerStates;
        QHash< quint8, QHash<quint8, bool> > _controlledPlayerStateIsNew;

    private:
        enum ResetState {WAIT_TIMEOUT, RESETTING, RESETED};
        QHash< quint8, QHash<quint8, ResetState> > _controlledPlayerResetState;
        QHash< quint8, QHash<quint8, QTime> > _controlledPlayerTimers;


    private:
        WRStation::CommandSerialInterface::ActuatorWorker* _actuatorWorker;
        WRStation::CommandSerialInterface::SerialWorker*   _serialWorker;
        QThread* _actuatorThread;
        QThread* _serialThread;


    private:
        bool _repeatCommands;


    public:
        CommandInterface(const QString& name, QObject* parent = 0);
        ~CommandInterface();


    public:
        WRStation::CommandSerialInterface::ActuatorWorker* actuatorWorker();
        WRStation::CommandSerialInterface::SerialWorker*   serialWorker();


    public slots:
        void setSpeed(quint8 teamNum, quint8 playerNum, float xSpeed, float ySpeed, float angularSpeed);
        void kickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power);
        void chipKickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power);
        void holdBall(quint8 teamNum, quint8 playerNum, bool enable);
        void kick(quint8 teamNum, quint8 playerNum, float power);
        void chipKick(quint8 teamNum, quint8 playerNum, float power);

        void checkPlayerState(quint8 teamNum, quint8 playerNum);
        void checkPlayerStates();
        void checkNewPlayerStates();

    signals:
        void playerStateChecked(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState) const;
        void playerStateCheckCompleted();

        void setSpeedReceived(quint8 teamNum, quint8 playerNum, float xSpeed, float ySpeed, float angularSpeed) const;
        void kickOnTouchReceived(quint8 teamNum, quint8 playerNum, bool enable, float power) const;
        void chipKickOnTouchReceived(quint8 teamNum, quint8 playerNum, bool enable, float power) const;
        void holdBallReceived(quint8 teamNum, quint8 playerNum, bool enable) const;
        void kickReceived(quint8 teamNum, quint8 playerNum, float power) const;
        void chipKickReceived(quint8 teamNum, quint8 playerNum, float power) const;


    public slots:
        void addControlledPlayer(quint8 teamNum, quint8 playerNum);
        void delControlledPlayer(quint8 teamNum, quint8 playerNum);
        void enableControlledPlayer(quint8 teamNum, quint8 playerNum, bool enable);

        void checkControlledPlayer(quint8 teamNum, quint8 playerNum) const;
        void checkEnabledPlayer(quint8 teamNum, quint8 playerNum) const;
        void checkControlledPlayers() const;

    signals:
        void controlledPlayerAdded(quint8 teamNum, quint8 playerNum) const;
        void controlledPlayerDeleted(quint8 teamNum, quint8 playerNum) const;
        void controlledPlayerChecked(quint8 teamNum, quint8 playerNum, bool controlled, bool enabled) const;
        void controlledPlayerEnabled(quint8 teamNum, quint8 playerNum, bool enabled) const;

    private:
        bool isPlayerControlled(quint8 teamNum, quint8 playerNum) const;
        bool isPlayerEnabled(quint8 teamNum, quint8 playerNum) const;


    public slots:
        void setPlayerAddress(quint8 teamNum, quint8 playerNum, const QByteArray& address);
        void checkPlayerAddress(quint8 teamNum, quint8 playerNum) const;
        void checkPlayerAddresses() const;

    signals:
        void playerAddressChanged(quint8 teamNum, quint8 playerNum, const QByteArray& address) const;
        void playerAddressChecked(quint8 teamNum, quint8 playerNum, const QByteArray& address) const;


    public slots:
        void setRepeatCommands(bool repeat);
        void checkRepeatCommands();

    signals:
        void repeatCommandsChanged(bool repeat);
        void repeatCommandsChecked(bool repeat);


    private:
        void createActuatorWorker(const QString& name);
        void connectActuatorWorkerSignals();
        void createSerialWorker();
        void connectSerialWorkerSignals();


    public slots:
        void printPlayerState(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState) const;
};


#endif
