/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Prevents multiple definitions
#ifndef WRSTATION_COMMANDSERIALINTERFACE_SERIALWORKER
#define WRSTATION_COMMANDSERIALINTERFACE_SERIALWORKER


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>
#include <QtSerialPort/QtSerialPort>


class WRStation::CommandSerialInterface::SerialWorker : public QObject {
    Q_OBJECT

    private:
        QSerialPort _serialPort;
        QTimer _timer;
        qint32 _timerInterval;
        WRStation::CommandSerialInterface::Protocol::Interface* _commandProtocol;
        QHash< quint8, QHash<quint8, quint32> >  _packetNonces;

        // Status to send
        QHash<quint8,QHash<quint8,WRStation::CommandPlayerState> > _statesToSend;
        bool _sendingCommands;


    public:
        SerialWorker(QObject* parent);
        virtual ~SerialWorker();


    public slots:
        void reset();

    signals:
        void hasReset();


    public slots:
        void serialPortClear();
        void serialPortClose();
        void serialPortFlush();
        void serialPortOpen(const QString& portName);
        void serialPortReset();
        void serialPortSetBaudRate(qint32 baudRate);
        void serialPortSetDataBits(QSerialPort::DataBits dataBits);
        void serialPortSetDataTerminalReady(bool set);
        void serialPortSetFlowControl(QSerialPort::FlowControl flowControl);
        void serialPortSetParity(QSerialPort::Parity parity);
        void serialPortSetReadBufferSize(qint64 size);
        void serialPortSetRequestToSend(bool set);
        void serialPortSetStopBits(QSerialPort::StopBits stopBits);

    signals:
        void serialPortCleared();
        void serialPortClosed();
        void serialPortOpened();
        void serialPortHasReset();
        void serialPortBaudRateChanged(qint32 baudRate);
        void serialPortDataBitsChanged(QSerialPort::DataBits dataBits);
        void serialPortDataTerminalReadyChanged(bool set);
        void serialPortError(QSerialPort::SerialPortError error);
        void serialPortFlowControlChanged(QSerialPort::FlowControl flow);
        void serialPortParityChanged(QSerialPort::Parity parity);
        void serialPortReadBufferSizeChanged(quint64 size);
        void serialPortRequestToSendChanged(bool set);
        void serialPortStopBitsChanged(QSerialPort::StopBits stopBits);
        void serialPortAboutToClose();
        void serialPortBytesWritten(qint64 bytes);
        void serialPortReadChannelFinished();
        void serialPortReadyRead();


    public slots:
        void checkSerialPortBaudRate() const;
        void checkSerialPortDataBits() const;
        void checkSerialPortError() const;
        void checkSerialPortErrorString() const;
        void checkSerialPortFlowControl() const;
        void checkSerialPortIsDataTerminalReady();
        void checkSerialPortIsOpen() const;
        void checkSerialPortIsRequestToSend();
        void checkSerialPortName() const;
        void checkSerialPortParity() const;
        void checkSerialPortReadBufferSize() const;
        void checkSerialPortStopBits() const;

    signals:
        void serialPortBaudRateChecked(qint32 baudRate) const;
        void serialPortDataBitsChecked(QSerialPort::DataBits dataBits) const;
        void serialPortErrorChecked(QSerialPort::SerialPortError error) const;
        void serialPortErrorStringChecked(const QString& errorString) const;
        void serialPortFlowControlChecked(QSerialPort::FlowControl flow) const;
        void serialPortIsDataTerminalReadyChecked(bool set) const;
        void serialPortIsOpenChecked(bool open) const;
        void serialPortIsRequestToSendChecked(bool set) const;
        void serialPortNameChecked(const QString& name) const;
        void serialPortParityChecked(QSerialPort::Parity parity) const;
        void serialPortReadBufferSizeChecked(qint64 size) const;
        void serialPortStopBitsChecked(QSerialPort::StopBits stopBits) const;


    public slots:
        void timerStart();
        void timerStop();
        void timerCheck() const;

    signals:
        void timerStarted() const;
        void timerStopped() const;
        void timerChecked(bool isRunning) const;


    private slots:
        void sendCommands();

    public slots:
        void addPlayerState(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState);
        void playersStateCompleted();


    signals:
        void requestPlayerStates();
        void commandsSent();
        void commandSent(quint8 teamNum, quint8 playerNum);
        void commandSendError(quint8 teamNum, quint8 playerNum, const QString& error);

    private:
        quint32 getNextNonce(quint8 teamNum, quint8 playerNum);


    private:
        void connectSerialPortSignals();
        void connectTimerSignals();
        void sendPlayerState(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState);
};


#endif
