/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Includes class header
#include <WRStation/CommandSerialInterface/GUI/actuatorworkergui.hh>

// Includes WRStation
#include <WRStation/wrstation.hh>
#include <WRStation/CommandSerialInterface/actuatorworker.hh>


// Includes Qt libraries
#include <QtCore/QtCore>
#include <QtGui/QtGui>
#include <QtWidgets/QtWidgets>


// Includes GEARSystem library
#include <GEARSystem/gearsystem.hh>


// Selects namespace
using namespace WRStation;
using namespace GEARSystem;


WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::ActuatorWorkerGUI(QWidget* parent, CommandSerialInterface::ActuatorWorker* actuatorWorker) : QWidget(parent) {
    // Sets the controlled worker
    _actuatorWorker = actuatorWorker;
    _isActuatorConnected = false;

    // Sets the object name
    this->setObjectName(this->metaObject()->className());

    // Setups the GUI
    this->setupGUI();

    // Connects the signals
    this->connectSignals();
}


void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setActuatorIsConnected(bool connected) {
    // Sets the flag and changes the button
    if (connected) {
        this->setActuatorConnected();
    } else {
        this->setActuatorDisconnected();
    }
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setActuatorConnecting() {
    // Sets the flags and changes the connect button
    _serverConnectButton->setEnabled(false);
    _serverIsConnectedLabel->setText(QString::fromUtf8("connecting"));
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setActuatorConnected() {
    // Sets the flags and changes the connect button
    _isActuatorConnected = true;
    _serverIsConnectedLabel->setText(QString::fromUtf8("connected"));
    _serverConnectButton->setText(QString::fromUtf8("Connected"));
//    _serverConnectButton->setText(QString::fromUtf8("Disconnect"));
//    _serverConnectButton->setEnabled(true);

    _serverAddressEdit->setEnabled(false);
    _serverPortEdit->setEnabled(false);
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setActuatorDisconnected() {
    // Sets the flags and changes the connect button
    _isActuatorConnected = false;
    _serverIsConnectedLabel->setText(QString::fromUtf8("not connected"));
    _serverConnectButton->setText(QString::fromUtf8("Connect"));
    _serverConnectButton->setEnabled(true);

    _serverAddressEdit->setEnabled(true);
    _serverPortEdit->setEnabled(true);
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setActuatorConnectFail(const QString &error) {
    // Sets the flags and changes the connect button
    _serverIsConnectedLabel->setText(error);
    _serverConnectButton->setEnabled(true);
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::onServerConnectButtonReleased() {
    // Calls the connect or disconnect methods
    if(_isActuatorConnected) {
        emit disconnectActuator();
    } else {
        emit connectActuator(_serverAddressEdit->text(), _serverPortEdit->value());
    }
}


void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::setupGUI() {
    // Creates the server connection layout
    QFormLayout* serverLayout = new QFormLayout();

    /// Create rows
    // Server address
    _serverAddressEdit = new QLineEdit(QString::fromUtf8("127.0.0.1"));

    // Server port
    _serverPortEdit = new QSpinBox();
    _serverPortEdit->setValue(0);
    _serverPortEdit->setMinimum(0);
    _serverPortEdit->setMaximum(65535);

    // Server connected label
    _serverIsConnectedLabel = new QLabel(QString::fromUtf8("not connected"));

    /// Add rows
    serverLayout->addRow(new QLabel(QString::fromUtf8("Server address:")), _serverAddressEdit);
    serverLayout->addRow(new QLabel(QString::fromUtf8("Server port:")), _serverPortEdit);
    serverLayout->addRow(new QLabel(QString::fromUtf8("Server connected:")), _serverIsConnectedLabel);

    /// Creates the button layout
    _serverConnectButton = new QPushButton(QString::fromUtf8("Connect"));
    QVBoxLayout* buttonLayout = new QVBoxLayout();
    buttonLayout->addWidget(_serverConnectButton);

    /// Creates the main layout
    QVBoxLayout* mainLayout = new QVBoxLayout();
    mainLayout->addLayout(serverLayout);
    mainLayout->addLayout(buttonLayout);

    // Sets the layout to the widget
    this->setLayout(mainLayout);
}

void WRStation::CommandSerialInterface::GUI::ActuatorWorkerGUI::connectSignals() {
    // Connects the worker signals
    QObject::connect(this, SIGNAL(connectActuator(QString,quint16)), _actuatorWorker, SLOT(connectActuator(QString,quint16)));
    QObject::connect(this, SIGNAL(disconnectActuator()), _actuatorWorker, SLOT(disconnectActuator()));

    QObject::connect(_actuatorWorker, SIGNAL(actuatorConnecting()), this, SLOT(setActuatorConnecting()));
    QObject::connect(_actuatorWorker, SIGNAL(actuatorConnected()), this, SLOT(setActuatorConnected()));
    QObject::connect(_actuatorWorker, SIGNAL(actuatorDisconnected()), this, SLOT(setActuatorDisconnected()));
    QObject::connect(_actuatorWorker, SIGNAL(actuatorConnectFail(QString)), this, SLOT(setActuatorConnectFail(QString)));
    QObject::connect(_actuatorWorker, SIGNAL(isActuatorConnected(bool)), this, SLOT(setActuatorIsConnected(bool)));

    // Connects the GUI signals
    QObject::connect(_serverConnectButton, SIGNAL(released()), this, SLOT(onServerConnectButtonReleased()));
}
