/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Includes class header
#include <WRStation/CommandSerialInterface/GUI/serialworkergui.hh>


// Includes WRStation
#include <WRStation/wrstation.hh>
#include <WRStation/CommandSerialInterface/serialworker.hh>


// Includes Qt libraries
#include <QtCore/QtCore>
#include <QtGui/QtGui>
#include <QtWidgets/QtWidgets>
#include <QtSerialPort/QtSerialPort>


// Selects namespace
using namespace WRStation;


WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::SerialWorkerGUI(QWidget* parent, CommandSerialInterface::SerialWorker* serialWorker) : QWidget(parent) {
    // Sets the controlled worker
    _serialWorker = serialWorker;

	_isSerialPortOpen = false;
	_isSerialPortRunning = false;

    // Sets the object name
    this->setObjectName(this->metaObject()->className());

    // Setups the GUI
    this->setupGUI();

    // Connects the signals
    this->connectSignals();

    _serialPortProtocolEdit->setEnabled(false);
}


void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortBaudRate(qint32 baudRate) {
    // Sets the baud rate
    _serialPortBaudEdit->setValue(baudRate);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortClosed() {
    // Sets the port as closed
    _isSerialPortOpen = false;
    _serialPortIsOpenLabel->setText(QString::fromUtf8("false"));

    // Disables the buttons
    _serialPortOpenButton->setText(QString::fromUtf8("Open"));
    _serialPortConfigureButton->setEnabled(false);
    _serialPortStartButton->setEnabled(false);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortDataBits(QSerialPort::DataBits dataBits) {
    // Sets the combo box
    int index;
    switch (dataBits) {
        case QSerialPort::Data5: {
            index = 0;
            break;
        }

        case QSerialPort::Data6: {
            index = 1;
            break;
        }

        case QSerialPort::Data7: {
            index = 2;
            break;
        }

        case QSerialPort::Data8: {
            index = 3;
            break;
        }

        default: {
            index = -1;
            break;
        }
    }
    _serialPortDataBitsEdit->setCurrentIndex(index);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortError(QSerialPort::SerialPortError) {
    // Asks for the error string
    emit checkSeriaPortErrorString();
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortErrorString(const QString& errorString) {
    // Sets the error string
    _serialPortErrorLabel->setText(errorString);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortFlowControl(QSerialPort::FlowControl flow) {
    // Sets the combo box
    int index;
    switch (flow) {
        case QSerialPort::NoFlowControl: {
            index = 0;
            break;
        }

        case QSerialPort::HardwareControl: {
            index = 1;
            break;
        }

        case QSerialPort::SoftwareControl: {
            index = 2;
            break;
        }

        default: {
            index = -1;
            break;
        }
    }
    _serialPortFlowControlEdit->setCurrentIndex(index);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortIsOpen(bool open) {
    // Sets the port information
    if (open) {
        this->setSerialPortOpen();
    }
    else {
        this->setSerialPortClosed();
    }
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortName(const QString& name) {
    // Sets the port name
    _serialPortNameEdit->setText(name);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortOpen() {
    // Sets the port as closed
    _isSerialPortOpen = true;
    _serialPortIsOpenLabel->setText(QString::fromUtf8("true"));

    // Disables the buttons
    _serialPortOpenButton->setText(QString::fromUtf8("Close"));
    _serialPortConfigureButton->setEnabled(true);
    _serialPortStartButton->setEnabled(true);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortParity(QSerialPort::Parity parity) {
    // Sets the combo box
    int index;
    switch (parity) {
        case QSerialPort::NoParity: {
            index = 0;
            break;
        }

        case QSerialPort::EvenParity: {
            index = 1;
            break;
        }

        case QSerialPort::OddParity: {
            index = 2;
            break;
        }

        case QSerialPort::SpaceParity: {
            index = 3;
            break;
        }

        case QSerialPort::MarkParity: {
            index = 4;
            break;
        }

        default: {
            index = -1;
            break;
        }
    }
    _serialPortParityEdit->setCurrentIndex(index);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setSerialPortStopBits(QSerialPort::StopBits stopBits) {
    // Sets the combo box
    int index;
    switch (stopBits) {
        case QSerialPort::OneStop: {
            index = 0;
            break;
        }

        case QSerialPort::OneAndHalfStop: {
            index = 1;
            break;
        }

        case QSerialPort::TwoStop: {
            index = 2;
            break;
        }

        default: {
            index = -1;
            break;
        }
    }
    _serialPortStopBitsEdit->setCurrentIndex(index);
}


void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setTimerIsRunning(bool isRunning) {
    // Sets the timer information
    if (isRunning) {
        this->setTimerStarted();
    }
    else {
        this->setTimerStopped();
    }
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setTimerStarted() {
    // Sets the port as started
    _isSerialPortRunning = true;
    _serialPortIsRunningLabel->setText(QString::fromUtf8("true"));

    // Changes the buttons
    _serialPortStartButton->setText(QString::fromUtf8("Stop"));
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setTimerStopped() {
    // Sets the port as stopped
    _isSerialPortRunning = false;
    _serialPortIsRunningLabel->setText(QString::fromUtf8("false"));

    // Changes the buttons
    _serialPortStartButton->setText(QString::fromUtf8("Start"));
}


void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::onSerialPortOpenButtonReleased() {
    // Opens or closes the serial port
    if (_isSerialPortOpen) {
        emit serialPortClose();
    } else {
        emit serialPortOpen(_serialPortNameEdit->text());
        this->onSerialPortConfigureButtonReleased(); // auto config
    }
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::onSerialPortConfigureButtonReleased() {
    // Baud rate
    emit serialPortSetBaudRate(_serialPortBaudEdit->value());

    // Data bits
    QString dataBitsStr = _serialPortDataBitsEdit->currentText();
    emit serialPortSetDataBits(this->str2databits(dataBitsStr));

    // Flow control
    QString flowControlStr = _serialPortFlowControlEdit->currentText();
    emit serialPortSetFlowControl(this->str2flowcontrol(flowControlStr));

    // Parity
    QString parityStr = _serialPortParityEdit->currentText();
    emit serialPortSetParity(this->str2parity(parityStr));

    // Stop bits
    QString stopBitsStr = _serialPortStopBitsEdit->currentText();
    emit serialPortSetStopBits(this->str2stopbits(stopBitsStr));
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::onSerialPortStartButtonReleased() {
    // Starts or stops the serial port
    if (_isSerialPortRunning) {
        _serialPortOpenButton->setEnabled(true);
        emit timerStop();
    } else {
        _serialPortOpenButton->setEnabled(false);
        emit timerStart();
    }
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::onSerialPortProtocolButtonReleased() {
    // TODO: Sets the protocol (Add available protocols list)
}


void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::setupGUI() {
    // Creates the serial port layout
    QFormLayout* serialPortLayout = new QFormLayout();

    _serialPortNameEdit = new QLineEdit(QString::fromUtf8("/dev/ttyUSB0"));

    _serialPortIsOpenLabel = new QLabel(QString::fromUtf8("false"));
    _serialPortIsRunningLabel = new QLabel(QString::fromUtf8("false"));
    _serialPortErrorLabel = new QLabel(QString::fromUtf8("no error"));

    _serialPortBaudEdit = new QSpinBox();
    _serialPortBaudEdit->setMinimum(1);
    _serialPortBaudEdit->setMaximum(3000000);
    _serialPortBaudEdit->setValue(1000000);

    _serialPortDataBitsEdit = new QComboBox();
    _serialPortDataBitsEdit->addItem(QString::fromUtf8("5"));
    _serialPortDataBitsEdit->addItem(QString::fromUtf8("6"));
    _serialPortDataBitsEdit->addItem(QString::fromUtf8("7"));
    _serialPortDataBitsEdit->addItem(QString::fromUtf8("8"));
    _serialPortDataBitsEdit->setCurrentIndex(3);

    _serialPortParityEdit = new QComboBox();
    _serialPortParityEdit->addItem(QString::fromUtf8("No"));
    _serialPortParityEdit->addItem(QString::fromUtf8("Even"));
    _serialPortParityEdit->addItem(QString::fromUtf8("Odd"));
    _serialPortParityEdit->addItem(QString::fromUtf8("Space"));
    _serialPortParityEdit->addItem(QString::fromUtf8("Mark"));
    _serialPortParityEdit->setCurrentIndex(0);

    _serialPortStopBitsEdit = new QComboBox();
    _serialPortStopBitsEdit->addItem(QString::fromUtf8("1"));
    _serialPortStopBitsEdit->addItem(QString::fromUtf8("1.5"));
    _serialPortStopBitsEdit->addItem(QString::fromUtf8("2"));
    _serialPortStopBitsEdit->setCurrentIndex(0);

    _serialPortFlowControlEdit = new QComboBox();
    _serialPortFlowControlEdit->addItem(QString::fromUtf8("No"));
    _serialPortFlowControlEdit->addItem(QString::fromUtf8("Hardware"));
    _serialPortFlowControlEdit->addItem(QString::fromUtf8("Software"));
    _serialPortFlowControlEdit->setCurrentIndex(0);

    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Path:")), _serialPortNameEdit);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Open:")), _serialPortIsOpenLabel);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Running:")), _serialPortIsRunningLabel);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Error:")), _serialPortErrorLabel);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Baud rate:")), _serialPortBaudEdit);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Data bits:")), _serialPortDataBitsEdit);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Parity:")), _serialPortParityEdit);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Stop bits:")), _serialPortStopBitsEdit);
    serialPortLayout->addRow(new QLabel(QString::fromUtf8("Flow control:")), _serialPortFlowControlEdit);

    // Creates the serial port button layout
    QHBoxLayout* serialPortButtonLayout = new QHBoxLayout();
    _serialPortOpenButton = new QPushButton(QString::fromUtf8("Open"));
    _serialPortConfigureButton = new QPushButton(QString::fromUtf8("Set configuration"));
    _serialPortConfigureButton->setEnabled(false);
    _serialPortStartButton = new QPushButton(QString::fromUtf8("Start"));
    _serialPortStartButton->setEnabled(false);
    serialPortButtonLayout->addWidget(_serialPortOpenButton);
    serialPortButtonLayout->addWidget(_serialPortConfigureButton);
    serialPortButtonLayout->addWidget(_serialPortStartButton);

    // Creates the protocol layout
    QFormLayout* protocolLayout = new QFormLayout();

    // TODO: Implement searching for available protocols
    _serialPortProtocolEdit = new QComboBox();
    _serialPortProtocolEdit->addItem(QString::fromUtf8("WRMagicV0"));
    _serialPortProtocolEdit->setCurrentIndex(0);

    _serialPortProtocolNameLabel = new QLabel(QString());
    _serialPortProtocolVersionLabel = new QLabel(QString());

    protocolLayout->addRow(new QLabel(QString::fromUtf8("Protocol:")), _serialPortProtocolEdit);
    protocolLayout->addRow(new QLabel(QString::fromUtf8("Name:")), _serialPortProtocolNameLabel);
    protocolLayout->addRow(new QLabel(QString::fromUtf8("Version:")), _serialPortProtocolVersionLabel);

    // Creates the protocol button layout
    QHBoxLayout* protocolButtonLayout = new QHBoxLayout();
    _serialPortProtocolButton = new QPushButton(QString::fromUtf8("Set protocol (TODO)"));
    protocolButtonLayout->addWidget(_serialPortProtocolButton);
    _serialPortProtocolButton->setEnabled(false); /// TODO: activate button

    // Creates the main layout
    QVBoxLayout* mainLayout = new QVBoxLayout();
    QFrame* spacing = new QFrame();
    spacing->setFrameShape(QFrame::HLine);

    // Assembles the layout tree
    mainLayout->addLayout(serialPortLayout);
    mainLayout->addLayout(serialPortButtonLayout);
    mainLayout->addWidget(spacing);
    mainLayout->addLayout(protocolLayout);
    mainLayout->addLayout(protocolButtonLayout);

    // Sets the layout to the widget
    this->setLayout(mainLayout);
}

void WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::connectSignals() {
    // Register QSerialPort types
    qRegisterMetaType<QSerialPort::DataBits>("QSerialPort::DataBits");
    qRegisterMetaType<QSerialPort::FlowControl>("QSerialPort::FlowControl");
    qRegisterMetaType<QSerialPort::Parity>("QSerialPort::Parity");
    qRegisterMetaType<QSerialPort::SerialPortError>("QSerialPort::SerialPortError");
    qRegisterMetaType<QSerialPort::StopBits>("QSerialPort::StopBits");

    // Connects the worker signals
    QObject::connect(this, SIGNAL(serialPortOpen(QString)), _serialWorker, SLOT(serialPortOpen(QString)));
    QObject::connect(this, SIGNAL(serialPortClose()), _serialWorker, SLOT(serialPortClose()));
    QObject::connect(this, SIGNAL(serialPortSetBaudRate(qint32)), _serialWorker, SLOT(serialPortSetBaudRate(qint32)));
    QObject::connect(this, SIGNAL(serialPortSetDataBits(QSerialPort::DataBits)), _serialWorker, SLOT(serialPortSetDataBits(QSerialPort::DataBits)));
    QObject::connect(this, SIGNAL(serialPortSetFlowControl(QSerialPort::FlowControl)), _serialWorker, SLOT(serialPortSetFlowControl(QSerialPort::FlowControl)));
    QObject::connect(this, SIGNAL(serialPortSetParity(QSerialPort::Parity)), _serialWorker, SLOT(serialPortSetParity(QSerialPort::Parity)));
    QObject::connect(this, SIGNAL(serialPortSetStopBits(QSerialPort::StopBits)), _serialWorker, SLOT(serialPortSetStopBits(QSerialPort::StopBits)));

    QObject::connect(_serialWorker, SIGNAL(serialPortBaudRateChanged(qint32)), this, SLOT(setSerialPortBaudRate(qint32)));
    QObject::connect(_serialWorker, SIGNAL(serialPortBaudRateChecked(qint32)), this, SLOT(setSerialPortBaudRate(qint32)));
    QObject::connect(_serialWorker, SIGNAL(serialPortClosed()), this, SLOT(setSerialPortClosed()));
    QObject::connect(_serialWorker, SIGNAL(serialPortDataBitsChanged(QSerialPort::DataBits)), this, SLOT(setSerialPortDataBits(QSerialPort::DataBits)));
    QObject::connect(_serialWorker, SIGNAL(serialPortDataBitsChecked(QSerialPort::DataBits)), this, SLOT(setSerialPortDataBits(QSerialPort::DataBits)));
    QObject::connect(_serialWorker, SIGNAL(serialPortError(QSerialPort::SerialPortError)), this, SLOT(setSerialPortError(QSerialPort::SerialPortError)));
    QObject::connect(_serialWorker, SIGNAL(serialPortErrorChecked(QSerialPort::SerialPortError)), this, SLOT(setSerialPortError(QSerialPort::SerialPortError)));
    QObject::connect(_serialWorker, SIGNAL(serialPortErrorStringChecked(QString)), this, SLOT(setSerialPortErrorString(QString)));
    QObject::connect(_serialWorker, SIGNAL(serialPortFlowControlChanged(QSerialPort::FlowControl)), this, SLOT(setSerialPortFlowControl(QSerialPort::FlowControl)));
    QObject::connect(_serialWorker, SIGNAL(serialPortFlowControlChecked(QSerialPort::FlowControl)), this, SLOT(setSerialPortFlowControl(QSerialPort::FlowControl)));
    QObject::connect(_serialWorker, SIGNAL(serialPortIsOpenChecked(bool)), this, SLOT(setSerialPortIsOpen(bool)));
    QObject::connect(_serialWorker, SIGNAL(serialPortNameChecked(QString)), this, SLOT(setSerialPortName(QString)));
    QObject::connect(_serialWorker, SIGNAL(serialPortOpened()), this, SLOT(setSerialPortOpen()));
    QObject::connect(_serialWorker, SIGNAL(serialPortParityChanged(QSerialPort::Parity)), this, SLOT(setSerialPortParity(QSerialPort::Parity)));
    QObject::connect(_serialWorker, SIGNAL(serialPortParityChecked(QSerialPort::Parity)), this, SLOT(setSerialPortParity(QSerialPort::Parity)));
    QObject::connect(_serialWorker, SIGNAL(serialPortStopBitsChanged(QSerialPort::StopBits)), this, SLOT(setSerialPortStopBits(QSerialPort::StopBits)));
    QObject::connect(_serialWorker, SIGNAL(serialPortStopBitsChecked(QSerialPort::StopBits)), this, SLOT(setSerialPortStopBits(QSerialPort::StopBits)));

    QObject::connect(this, SIGNAL(timerStart()), _serialWorker, SLOT(timerStart()));
    QObject::connect(this, SIGNAL(timerStop()), _serialWorker, SLOT(timerStop()));

    QObject::connect(_serialWorker, SIGNAL(timerChecked(bool)), this, SLOT(setTimerIsRunning(bool)));
    QObject::connect(_serialWorker, SIGNAL(timerStarted()), this, SLOT(setTimerStarted()));
    QObject::connect(_serialWorker, SIGNAL(timerStopped()), this, SLOT(setTimerStopped()));

    QObject::connect(this, SIGNAL(checkSeriaPortErrorString()), _serialWorker, SLOT(checkSerialPortErrorString()));

    // Connects the GUI signals
    QObject::connect(_serialPortOpenButton, SIGNAL(released()), this, SLOT(onSerialPortOpenButtonReleased()));
    QObject::connect(_serialPortConfigureButton, SIGNAL(released()), this, SLOT(onSerialPortConfigureButtonReleased()));
    QObject::connect(_serialPortStartButton, SIGNAL(released()), this, SLOT(onSerialPortStartButtonReleased()));
    QObject::connect(_serialPortProtocolButton, SIGNAL(released()), this, SLOT(onSerialPortProtocolButtonReleased()));
}

QSerialPort::DataBits WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::str2databits(QString dataBitsStr) {
    if(dataBitsStr=="5") return QSerialPort::Data5;
    if(dataBitsStr=="6") return QSerialPort::Data6;
    if(dataBitsStr=="7") return QSerialPort::Data7;
    if(dataBitsStr=="8") return QSerialPort::Data8;
    return QSerialPort::UnknownDataBits;
}

QSerialPort::FlowControl WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::str2flowcontrol(QString flowControlStr) {
    if(flowControlStr=="No") return QSerialPort::NoFlowControl;
    if(flowControlStr=="Hardware") return QSerialPort::HardwareControl;
    if(flowControlStr=="Software") return QSerialPort::SoftwareControl;
    return QSerialPort::UnknownFlowControl;
}

QSerialPort::Parity WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::str2parity(QString parityStr) {
    if(parityStr=="No") return QSerialPort::NoParity;
    if(parityStr=="Even") return QSerialPort::EvenParity;
    if(parityStr=="Odd") return QSerialPort::OddParity;
    if(parityStr=="Space") return QSerialPort::SpaceParity;
    if(parityStr=="Mark") return QSerialPort::MarkParity;
    return QSerialPort::UnknownParity;
}

QSerialPort::StopBits WRStation::CommandSerialInterface::GUI::SerialWorkerGUI::str2stopbits(QString stopBitsStr) {
    if(stopBitsStr=="1") return QSerialPort::OneStop;
    if(stopBitsStr=="1") return QSerialPort::OneAndHalfStop;
    if(stopBitsStr=="1") return QSerialPort::TwoStop;
    return QSerialPort::UnknownStopBits;
}
