/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Includes class header
#include <WRStation/CommandSerialInterface/Protocol/wrmagiccommandv0.hh>


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>


// Selects namespace
using namespace WRStation;

QString WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0::protocolName() const {
    // Returns the protocol name
    return(QString::fromUtf8("WRMagicCommandV0"));
}

quint8 WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0::protocolCode() const {
    // Returns the protocol code
    return(1);
}

quint8 WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0::protocolVersion() const {
    // Returns the protocol version
    return(1);
}


QByteArray WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0::encodeCommand(const WRStation::CommandPlayerState& playerState, quint32 nonce) {
    // Gets the address
    QByteArray command = playerState.playerAddress().left(5);

    // Allocates the command raw data vector and inserts the nonce
    char commandData[26];
    memcpy((void*) commandData, (void*) &nonce, 4);

    // Inserts the speeds
    float speed;
    speed = playerState.xSpeed();
    memcpy((void*) (commandData+4), (void*) &speed, 4);
    speed = playerState.ySpeed();
    memcpy((void*) (commandData+8), (void*) &speed, 4);
    speed = playerState.angularSpeed();
    memcpy((void*) (commandData+12), (void*) &speed, 4);

    // Inserts a 0 for backward compatibility
    speed = 0.0;
    memcpy((void*) (commandData+16), (void*) &speed, 4);

    // Inserts the hold ball information
    speed = playerState.holdBallEnable() ? 1.0 : 0.0;
    memcpy((void*) (commandData+20), (void*) &speed, 4);

    // Inserts kick information
    quint8 kickType, kickPower;
    bool isKicking     = playerState.kickOnTouchEnable();
    bool isChipKicking = playerState.chipKickOnTouchEnable();

    if (isKicking) {
        kickType = 3;
        kickPower = playerState.kickOnTouchPower()*10;
    }

    else {
        if (isChipKicking) {
            kickType = 4;
            kickPower = playerState.chipKickOnTouchPower()*10;
        }

        else {
            kickType  = 0;
            kickPower = 0;
        }
    }

    memcpy((void*) (commandData+24), (void*) &kickType, 1);
    memcpy((void*) (commandData+25), (void*) &kickPower, 1);

    // Converts the data packet to QByteArray and inserts the player address
    command.append(QByteArray(commandData, 26));

    // Returns the command
    return(command);
}

WRStation::CommandPlayerState WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0::decodeCommand(const QByteArray& rawData, quint32& nonce) {
    // Gets the player address
    CommandPlayerState playerState;
    playerState.setPlayerAddress(rawData.left(5));

    // Creates a raw data vector from the QByteArray
    const char* raw = rawData.data();

    // Gets the nonce
    memcpy((void*) &nonce, (void*) raw, 4);

    // Gets the speeds
    float speed;
    memcpy((void*) &speed, (void*) (raw+4), 4);
    playerState.setXSpeed(speed);
    memcpy((void*) &speed, (void*) (raw+8), 4);
    playerState.setYSpeed(speed);
    memcpy((void*) &speed, (void*) (raw+12), 4);
    playerState.setAngularSpeed(speed);

    // Gets the hold ball information
    memcpy((void*) &speed, (void*) (raw+20), 4);
    if (!qIsNull(speed)) {
        playerState.holdBall(true);
    }

    // Gets kick data
    quint8 kickType, kickPower;
    memcpy((void*) &kickType, (void*) (raw+24), 1);
    memcpy((void*) &kickPower, (void*) (raw+25), 1);

    if (kickType == 3) {
        playerState.kickOnTouch(true, kickPower/10.0);
    }
    else if (kickType == 4) {
        playerState.chipKickOnTouch(true, kickPower/10.0);
    }

    // Returns the player state
    return(playerState);
}
