/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Includes class header
#include <WRStation/CommandSerialInterface/actuatorworker.hh>


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>


// Includes GEARSystem library
#include <GEARSystem/gearsystem.hh>

#include <iostream>


// Selects namespace
using namespace WRStation;
using namespace GEARSystem;


WRStation::CommandSerialInterface::ActuatorWorker::ActuatorWorker(QObject* parent, const QString& name) : QObject(parent), Actuator(name) {
    // Sets the object name
    this->setObjectName(this->metaObject()->className());
}


void WRStation::CommandSerialInterface::ActuatorWorker::reset() {
    // Disconnects from the GEARSystem server
    this->disconnectActuator();

    // Clears the controlled players list
    _controlledPlayers.clear();

    // Emits a signal
    emit hasReset();
}


void WRStation::CommandSerialInterface::ActuatorWorker::connectActuator(const QString& address, uint16 port) {

    // Emits connecting signal
    emit actuatorConnecting();

    // Connects to the GEARSystem server
    Actuator* actuator = static_cast<Actuator*>(this);
    bool ok = actuator->connect(address, port);

    // Emits a signal
    if (ok) {
        emit actuatorConnected();
    } else {
        emit actuatorConnectFail(QString("failed to connect"));
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::disconnectActuator() {
    // Disconnects from the GEARSystem server and emits a signal
    Actuator* actuator = static_cast<Actuator*>(this);
    actuator->disconnect();
    emit actuatorDisconnected();
}

void WRStation::CommandSerialInterface::ActuatorWorker::checkIsActuatorConnected() const {
    // Verifies if the actuator is connected and emits a signal
    const Actuator* actuator = static_cast<const Actuator*>(this);
    bool connected = actuator->isConnected();
    emit isActuatorConnected(connected);
}


void WRStation::CommandSerialInterface::ActuatorWorker::setSpeed(quint8 teamNum, quint8 playerNum, float xSpeed, float ySpeed, float angularSpeed) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit setSpeedReceived(teamNum, playerNum, xSpeed, ySpeed, angularSpeed);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::kickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit kickOnTouchReceived(teamNum, playerNum, enable, power);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::chipKickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit chipKickOnTouchReceived(teamNum, playerNum, enable, power);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::holdBall(quint8 teamNum, quint8 playerNum, bool enable) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit holdBallReceived(teamNum, playerNum, enable);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::kick(quint8 teamNum, quint8 playerNum, float power) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit kickReceived(teamNum, playerNum, power);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::chipKick(quint8 teamNum, quint8 playerNum, float power) {
    teamNum = 0;

    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit chipKickReceived(teamNum, playerNum, power);
    }
}


void WRStation::CommandSerialInterface::ActuatorWorker::addControlledPlayer(quint8 teamNum, quint8 playerNum) {
    // Adds the player to the controlled list and enables it
    _controlledPlayers[teamNum][playerNum] = true;
}

void WRStation::CommandSerialInterface::ActuatorWorker::delControlledPlayer(quint8 teamNum, quint8 playerNum) {
    // Removes the player from the controlled list and deletes the player state set-points
    if (_controlledPlayers.contains(teamNum)) {
        _controlledPlayers[teamNum].remove(playerNum);

        // Clears the team if it was the last player
        if (_controlledPlayers[teamNum].size() == 0) {
            _controlledPlayers.remove(teamNum);
        }
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::setControlledPlayer(quint8 teamNum, quint8 playerNum, bool controlled, bool enabled) {
    // Sets the if player is controlled
    if (controlled) {
        this->addControlledPlayer(teamNum, playerNum);
        this->enableControlledPlayer(teamNum, playerNum, enabled);
    }
    else {
        this->delControlledPlayer(teamNum, playerNum);
    }
}

void WRStation::CommandSerialInterface::ActuatorWorker::enableControlledPlayer(quint8 teamNum, quint8 playerNum, bool enable) {
    // Enables or disables the player
    if (this->isPlayerControlled(teamNum, playerNum)) {
        _controlledPlayers[teamNum][playerNum] = enable;
    }
}

bool WRStation::CommandSerialInterface::ActuatorWorker::isPlayerControlled(quint8 teamNum, quint8 playerNum) const {
    // Verifies if the player is in the controlled list
    return(_controlledPlayers.value(teamNum).contains(playerNum));
}

bool WRStation::CommandSerialInterface::ActuatorWorker::isPlayerEnabled(quint8 teamNum, quint8 playerNum) const {
    // Returns the "is enabled" flag of the player
    return(_controlledPlayers.value(teamNum).value(playerNum, false));
}
