/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

// Includes class header
#include <WRStation/CommandSerialInterface/commandinterface.hh>
#include <WRStation/CommandSerialInterface/actuatorworker.hh>
#include <WRStation/CommandSerialInterface/serialworker.hh>


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>

#include <iostream>
#define PLAYER_RESET_TIMEOUT 100
#define PLAYER_RESET_TIME 1000

// Selects namespace
using namespace WRStation;


WRStation::CommandSerialInterface::CommandInterface::CommandInterface(const QString& name, QObject* parent) : QObject(parent) {
    // Sets the object name
    this->setObjectName(this->metaObject()->className());

    // Registers Qt's meta types
    qRegisterMetaType<WRStation::CommandPlayerState>("WRStation::CommandPlayerState");

    // Creates the actuator and serial workers
    this->createActuatorWorker(name);
    this->createSerialWorker();

    // Testing stuff
    char a[] = {'W','R','M','0','0'};
    QByteArray addr(a, 5);
    for(int i=0; i<12; i++) {
        this->addControlledPlayer(0, i);
        addr.replace(4, 1, QByteArray::number(i, 16).toUpper().right(1));
        _controlledPlayerStates[0][i].setPlayerAddress(addr);
    }
}

WRStation::CommandSerialInterface::CommandInterface::~CommandInterface() {
    // Stops the threads execution loop
    _actuatorThread->quit();
    _serialThread->quit();
    _actuatorThread->wait();
    _serialThread->wait();
}


WRStation::CommandSerialInterface::ActuatorWorker* WRStation::CommandSerialInterface::CommandInterface::actuatorWorker() {
    // Returns the actuator worker
    return(_actuatorWorker);
}

WRStation::CommandSerialInterface::SerialWorker* WRStation::CommandSerialInterface::CommandInterface::serialWorker() {
    // Returns the serial worker
    return(_serialWorker);
}


void WRStation::CommandSerialInterface::CommandInterface::setSpeed(quint8 teamNum, quint8 playerNum, float xSpeed, float ySpeed, float angularSpeed) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        _controlledPlayerStates[teamNum][playerNum].setSpeed(xSpeed, ySpeed, angularSpeed);
        _controlledPlayerStateIsNew[teamNum][playerNum] = true;
        emit setSpeedReceived(teamNum, playerNum, xSpeed, ySpeed, angularSpeed);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::kickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        _controlledPlayerStates[teamNum][playerNum].kickOnTouch(enable, enable ? power : 0.0);
        _controlledPlayerStateIsNew[teamNum][playerNum] = true;
        emit kickOnTouchReceived(teamNum, playerNum, enable, power);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::chipKickOnTouch(quint8 teamNum, quint8 playerNum, bool enable, float power) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        _controlledPlayerStates[teamNum][playerNum].chipKickOnTouch(enable, enable ? power : 0.0);
        _controlledPlayerStateIsNew[teamNum][playerNum] = true;
        emit chipKickOnTouchReceived(teamNum, playerNum, enable, power);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::holdBall(quint8 teamNum, quint8 playerNum, bool enable) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        _controlledPlayerStates[teamNum][playerNum].holdBall(enable);
        _controlledPlayerStateIsNew[teamNum][playerNum] = true;
        emit holdBallReceived(teamNum, playerNum, enable);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::kick(quint8 teamNum, quint8 playerNum, float power) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit kickReceived(teamNum, playerNum, power);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::chipKick(quint8 teamNum, quint8 playerNum, float power) {
    // Verifies if the player is enabled and emits the signal
    if (this->isPlayerEnabled(teamNum, playerNum)) {
        emit chipKickReceived(teamNum, playerNum, power);
    }
}


void WRStation::CommandSerialInterface::CommandInterface::checkPlayerState(quint8 teamNum, quint8 playerNum) {
    // Verifies if the player is controlled
    bool isControlled = this->isPlayerControlled(teamNum, playerNum);

    // If true, emits the singal
    if (isControlled) {
        // Emits the signal only if the player is enabled
        bool isEnabled = this->isPlayerEnabled(teamNum, playerNum);
        if (isEnabled) {
            _controlledPlayerStateIsNew[teamNum][playerNum] = false;
            emit playerStateChecked(teamNum, playerNum, _controlledPlayerStates.value(teamNum).value(playerNum));
        }

        // Emits another signal if the player isn't enabled
        else {
            emit controlledPlayerChecked(teamNum, playerNum, isControlled, isEnabled);
        }
    }
}

void WRStation::CommandSerialInterface::CommandInterface::checkPlayerStates() {
    // Runs the controlled teams list
    QHashIterator< quint8, QHash<quint8, CommandPlayerState> > teamIterator(_controlledPlayerStates);
    while (teamIterator.hasNext()) {

        // Runs the controlled players list
        (void) teamIterator.next();
        QHashIterator<quint8, CommandPlayerState> playerIterator(teamIterator.value());
        while (playerIterator.hasNext()) {

            // Gets the team and the player numbers
            (void) playerIterator.next();
            quint8 teamNum   = teamIterator.key();
            quint8 playerNum = playerIterator.key();

            // Emits the singal for each controlled and enabled player
            bool isEnabled = this->isPlayerEnabled(teamNum, playerNum);
            if (isEnabled) {
                _controlledPlayerStateIsNew[teamNum][playerNum] = false;
                emit playerStateChecked(teamNum, playerNum, playerIterator.value());
            }
        }
    }

    emit playerStateCheckCompleted();
}

void WRStation::CommandSerialInterface::CommandInterface::checkNewPlayerStates() {
    // Runs the controlled teams list
    QHashIterator< quint8, QHash<quint8, CommandPlayerState> > teamIterator(_controlledPlayerStates);
    while (teamIterator.hasNext()) {

        // Runs the controlled players list
        (void) teamIterator.next();
        QHashIterator<quint8, CommandPlayerState> playerIterator(teamIterator.value());
        while (playerIterator.hasNext()) {

            // Gets the team and the player numbers
            (void) playerIterator.next();
            quint8 teamNum   = teamIterator.key();
            quint8 playerNum = playerIterator.key();

            // Emits the singal for each controlled and enabled player with new state
            bool isEnabled = this->isPlayerEnabled(teamNum, playerNum);
            if (isEnabled) {
                bool isNew = _controlledPlayerStateIsNew.value(teamNum).value(playerNum);
                if (isNew) {
                    _controlledPlayerTimers[teamNum][playerNum].start();
                    _controlledPlayerResetState[teamNum][playerNum] = WAIT_TIMEOUT;

                    _controlledPlayerStateIsNew[teamNum][playerNum] = false;
                    emit playerStateChecked(teamNum, playerNum, playerIterator.value());
                } else { // check player reset


                    switch(_controlledPlayerResetState[teamNum][playerNum]) {
                        default:
                        case WAIT_TIMEOUT: {
                            // Check reset timer
                            int ms = _controlledPlayerTimers[teamNum][playerNum].elapsed();
                            if(ms > PLAYER_RESET_TIMEOUT) {
                                // Reset player
                                _controlledPlayerStates[teamNum][playerNum].setSpeed(0.0, 0.0, 0.0);
                                _controlledPlayerStates[teamNum][playerNum].enableKickOnTouch(false);
                                _controlledPlayerStates[teamNum][playerNum].setKickOnTouchPower(0.0);
                                _controlledPlayerStates[teamNum][playerNum].enableChipKickOnTouch(false);
                                _controlledPlayerStates[teamNum][playerNum].setChipKickOnTouchPower(0.0);
                                _controlledPlayerStates[teamNum][playerNum].enableHoldBall(false);

                                // Start sending reset
                                _controlledPlayerTimers[teamNum][playerNum].restart();
                                _controlledPlayerResetState[teamNum][playerNum] = RESETTING;
                            }
                        } break;

                        case RESETTING: {
                            // Emit reset
                            emit playerStateChecked(teamNum, playerNum, _controlledPlayerStates[teamNum][playerNum]);

                            // Check time to stop reseting
                            int ms = _controlledPlayerTimers[teamNum][playerNum].elapsed();
                            if(ms > PLAYER_RESET_TIME) {
                                std::cout << "[COMMAND INTERFACE] Player #" << (int)playerNum << " reseted!\n";
                                _controlledPlayerResetState[teamNum][playerNum] = RESETED;
                            }
                        } break;

                        case RESETED: {
                            // Do nothing (reseted!)

                        } break;
                    }

                }
            }
        }
    }

    emit playerStateCheckCompleted();
}


void WRStation::CommandSerialInterface::CommandInterface::addControlledPlayer(quint8 teamNum, quint8 playerNum) {
    // Adds the player to the controlled list and enables it
    _controlledPlayers[teamNum][playerNum] = true;

    // Initializes the player state set-points
    _controlledPlayerStates[teamNum][playerNum];
    _controlledPlayerStates[teamNum][playerNum].setTeamNumber(teamNum);
    _controlledPlayerStates[teamNum][playerNum].setPlayerNumber(playerNum);
    _controlledPlayerStateIsNew[teamNum][playerNum] = true;

    // Emits a singal
    emit controlledPlayerAdded(teamNum, playerNum);
    emit controlledPlayerChecked(teamNum, playerNum, true, true);
}

void WRStation::CommandSerialInterface::CommandInterface::delControlledPlayer(quint8 teamNum, quint8 playerNum) {
    // Removes the player from the controlled list and deletes the player state set-points
    if (_controlledPlayers.contains(teamNum)) {
        _controlledPlayers[teamNum].remove(playerNum);
        _controlledPlayerStates[teamNum].remove(playerNum);
        _controlledPlayerStateIsNew[teamNum].remove(playerNum);

        // Clears the team if it was the last player
        if (_controlledPlayers[teamNum].size() == 0) {
            _controlledPlayers.remove(teamNum);
            _controlledPlayerStates.remove(teamNum);
            _controlledPlayerStateIsNew.remove(teamNum);
        }

        // Emits a singal
        emit controlledPlayerDeleted(teamNum, playerNum);
        emit controlledPlayerChecked(teamNum, playerNum, false, false);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::enableControlledPlayer(quint8 teamNum, quint8 playerNum, bool enable) {
    // Enables or disables the player
    if (this->isPlayerControlled(teamNum, playerNum)) {
        _controlledPlayers[teamNum][playerNum] = enable;

        if (!enable) {
            _controlledPlayerStates[teamNum][playerNum].clearState();
            _controlledPlayerStateIsNew[teamNum][playerNum] = false;
        }

        // Emits a singal
        emit controlledPlayerEnabled(teamNum, playerNum, enable);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::checkControlledPlayer(quint8 teamNum, quint8 playerNum) const {
    // Verifies if the player is controlled
    bool isControlled = this->isPlayerControlled(teamNum, playerNum);
    bool isEnabled    = this->isPlayerEnabled(teamNum, playerNum);

    // Emits the signal
    emit controlledPlayerChecked(teamNum, playerNum, isControlled, isEnabled);
}

void WRStation::CommandSerialInterface::CommandInterface::checkEnabledPlayer(quint8 teamNum, quint8 playerNum) const {
    // Verifies if the player is enabled
    bool isEnabled = this->isPlayerEnabled(teamNum, playerNum);

    // Emits the signal
    emit controlledPlayerEnabled(teamNum, playerNum, isEnabled);
}

void WRStation::CommandSerialInterface::CommandInterface::checkControlledPlayers() const {
    // Runs the controlled teams list
    QHashIterator< quint8, QHash<quint8, bool> > teamIterator(_controlledPlayers);
    while (teamIterator.hasNext()) {

        // Runs the controlled players list
        (void) teamIterator.next();
        QHashIterator<quint8, bool> playerIterator(teamIterator.value());
        while (playerIterator.hasNext()) {

            // Gets the team and the player numbers
            (void) playerIterator.next();
            quint8 teamNum   = teamIterator.key();
            quint8 playerNum = playerIterator.key();
            bool   isEnabled = playerIterator.value();

            // Emits the singal for each controlled player
            emit controlledPlayerChecked(teamNum, playerNum, true, isEnabled);
        }
    }
}

bool WRStation::CommandSerialInterface::CommandInterface::isPlayerControlled(quint8 teamNum, quint8 playerNum) const {
    // Verifies if the player is in the controlled list
    return(_controlledPlayers.value(teamNum).contains(playerNum));
}

bool WRStation::CommandSerialInterface::CommandInterface::isPlayerEnabled(quint8 teamNum, quint8 playerNum) const {
    // Returns the "is enabled" flag of the player
    return(_controlledPlayers.value(teamNum).value(playerNum, false));
}


void WRStation::CommandSerialInterface::CommandInterface::setPlayerAddress(quint8 teamNum, quint8 playerNum, const QByteArray& address) {
    // Sets the player address and emits a signal
    if (this->isPlayerControlled(teamNum, playerNum)) {
        _controlledPlayerStates[teamNum][playerNum].setPlayerAddress(address);
        emit playerAddressChanged(teamNum, playerNum, address);
    }
}

void WRStation::CommandSerialInterface::CommandInterface::checkPlayerAddress(quint8 teamNum, quint8 playerNum) const {
    // Emits a signal with the player address
    emit playerAddressChecked(teamNum, playerNum, _controlledPlayerStates.value(teamNum).value(playerNum).playerAddress());
}

void WRStation::CommandSerialInterface::CommandInterface::checkPlayerAddresses() const {
    // Runs the controlled teams list
    QHashIterator< quint8, QHash<quint8, CommandPlayerState> > teamIterator(_controlledPlayerStates);
    while (teamIterator.hasNext()) {

        // Runs the controlled players list
        (void) teamIterator.next();
        QHashIterator<quint8, CommandPlayerState> playerIterator(teamIterator.value());
        while (playerIterator.hasNext()) {

            // Gets the team and the player numbers
            (void) playerIterator.next();
            quint8 teamNum   = teamIterator.key();
            quint8 playerNum = playerIterator.key();

            // Emits the singal for each controlled and enabled player
            bool isEnabled = this->isPlayerEnabled(teamNum, playerNum);
            if (isEnabled) {
                emit playerAddressChecked(teamNum, playerNum, playerIterator.value().playerAddress());
            }
        }
    }
}


void WRStation::CommandSerialInterface::CommandInterface::setRepeatCommands(bool repeat) {
    // Sets the repeat commands flag and the proper connections
    _repeatCommands = repeat;

    if (_repeatCommands) {
        QObject::disconnect(_serialWorker, &WRStation::CommandSerialInterface::SerialWorker::requestPlayerStates, this, &WRStation::CommandSerialInterface::CommandInterface::checkNewPlayerStates);
        QObject::connect(_serialWorker, &WRStation::CommandSerialInterface::SerialWorker::requestPlayerStates, this, &WRStation::CommandSerialInterface::CommandInterface::checkPlayerStates);
    }
    else {
        QObject::disconnect(_serialWorker, &WRStation::CommandSerialInterface::SerialWorker::requestPlayerStates, this, &WRStation::CommandSerialInterface::CommandInterface::checkPlayerStates);
        QObject::connect(_serialWorker, &WRStation::CommandSerialInterface::SerialWorker::requestPlayerStates, this, &WRStation::CommandSerialInterface::CommandInterface::checkNewPlayerStates);
    }

    // Emits a signal with the new flag
    emit repeatCommandsChanged(_repeatCommands);
}

void WRStation::CommandSerialInterface::CommandInterface::checkRepeatCommands() {
    // Emits a signal with the flag
    emit repeatCommandsChecked(_repeatCommands);
}


void WRStation::CommandSerialInterface::CommandInterface::createActuatorWorker(const QString& name) {
    // Creates the actuator interface thread and its worker
    _actuatorThread = new QThread(this);
    QString objName = this->metaObject()->className();
    objName.append(QString::fromUtf8("ActuatorThread"));
    _actuatorThread->setObjectName(objName);

    _actuatorWorker = new CommandSerialInterface::ActuatorWorker(0, name);
    _actuatorWorker->moveToThread(_actuatorThread);

    // Connects their signals
    this->connectActuatorWorkerSignals();

    // Starts the thread
    _actuatorThread->start();
}

void WRStation::CommandSerialInterface::CommandInterface::connectActuatorWorkerSignals() {
    // Connects the controlled players management signals
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::controlledPlayerAdded,   _actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::addControlledPlayer, Qt::BlockingQueuedConnection);
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::controlledPlayerDeleted, _actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::delControlledPlayer, Qt::BlockingQueuedConnection);
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::controlledPlayerChecked, _actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::setControlledPlayer, Qt::BlockingQueuedConnection);
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::controlledPlayerEnabled, _actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::enableControlledPlayer, Qt::BlockingQueuedConnection);

    // Connects command signals
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::setSpeedReceived,        this, &WRStation::CommandSerialInterface::CommandInterface::setSpeed, Qt::BlockingQueuedConnection);
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::kickOnTouchReceived,     this, &WRStation::CommandSerialInterface::CommandInterface::kickOnTouch, Qt::BlockingQueuedConnection);
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::chipKickOnTouchReceived, this, &WRStation::CommandSerialInterface::CommandInterface::chipKickOnTouch, Qt::BlockingQueuedConnection);
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::holdBallReceived,        this, &WRStation::CommandSerialInterface::CommandInterface::holdBall, Qt::BlockingQueuedConnection);
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::kickReceived,            this, &WRStation::CommandSerialInterface::CommandInterface::kick, Qt::BlockingQueuedConnection);
    QObject::connect(_actuatorWorker, &WRStation::CommandSerialInterface::ActuatorWorker::chipKickReceived,        this, &WRStation::CommandSerialInterface::CommandInterface::chipKick, Qt::BlockingQueuedConnection);
}

void WRStation::CommandSerialInterface::CommandInterface::createSerialWorker() {
    // Creates the serial interface thread and its worker
    _serialThread = new QThread(this);
    QString objName = this->metaObject()->className();
    objName.append(QString::fromUtf8("SerialThread"));
    _serialThread->setObjectName(objName);

    _serialWorker = new CommandSerialInterface::SerialWorker(0);
    _serialWorker->moveToThread(_serialThread);

    // Connects their signals
    this->connectSerialWorkerSignals();

    // Starts the thread
    _serialThread->start();
}

void WRStation::CommandSerialInterface::CommandInterface::connectSerialWorkerSignals() {
    // Connects the states sharing signals
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::playerStateChecked, _serialWorker, &WRStation::CommandSerialInterface::SerialWorker::addPlayerState, Qt::BlockingQueuedConnection);
    QObject::connect(this, &WRStation::CommandSerialInterface::CommandInterface::playerStateCheckCompleted, _serialWorker, &WRStation::CommandSerialInterface::SerialWorker::playersStateCompleted, Qt::BlockingQueuedConnection);

    this->setRepeatCommands(false); /// TODO: move to GUI
}


void WRStation::CommandSerialInterface::CommandInterface::printPlayerState(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState) const {
    qDebug() << QString().sprintf("Team #%d - Player #%d:", teamNum, playerNum);
    qDebug() << QString().sprintf(" - Speed: %.1f , %.1f , %.1f", playerState.xSpeed(), playerState.ySpeed(), playerState.angularSpeed());
    qDebug() << QString().sprintf(" - Hold ball: %d", playerState.holdBallEnable());
    qDebug() << QString().sprintf(" - KickOnTouch: %d , %.1f", playerState.kickOnTouchEnable(), playerState.kickOnTouchPower());
    qDebug() << QString().sprintf(" - ChipKickOnTouch: %d , %.1f", playerState.chipKickOnTouchEnable(), playerState.chipKickOnTouchPower());
}
