/**
  * Warthog Robotics
  * University of Sao Paulo (USP) at Sao Carlos
  * http://www.warthog.sc.usp.br/
  * This file is part of the WRStation project
  *
  * Copyright (C) 2017 Warthog Robotics
  *
  * This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
  * the Free Software Foundation, either version 3 of the License, or
  * (at your option) any later version.
  *
  * This program is distributed in the hope that it will be useful,
  * but WITHOUT ANY WARRANTY; without even the implied warranty of
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  * GNU General Public License for more details.
  *
  * You should have received a copy of the GNU General Public License
  * along with this program. If not, see <http://www.gnu.org/licenses/>.
  **/

#pragma GCC diagnostic ignored "-Wunused-parameter"

// Includes class header
#include <WRStation/CommandSerialInterface/serialworker.hh>


// Includes WRStation
#include <WRStation/wrstation.hh>


// Includes Qt libraries
#include <QtCore/QtCore>
#include <QtSerialPort/QtSerialPort>

#include <iostream>

// Selects namespace
using namespace WRStation;


WRStation::CommandSerialInterface::SerialWorker::SerialWorker(QObject* parent) : QObject(parent), _serialPort(this), _timer(this) {
    // Sets the object name
    this->setObjectName(this->metaObject()->className());

    // Initializes the timer interval with a default value (60 Hz) and connects its signals
    _timerInterval = 1000/60;
    _timer.setTimerType(Qt::PreciseTimer);
    this->connectTimerSignals();

    // Connects the serial port signals
    this->connectSerialPortSignals();

    // Command protocol
    _commandProtocol = new WRStation::CommandSerialInterface::Protocol::WRMagicCommandV0();

    // Initialization
    _statesToSend.insert(0, QHash<uint8,CommandPlayerState>());
    _sendingCommands = false;

    std::cout.setf(std::ios::fixed, std::ios::floatfield);
    std::cout.precision(2);
}

WRStation::CommandSerialInterface::SerialWorker::~SerialWorker() {
    delete _commandProtocol;
}

void WRStation::CommandSerialInterface::SerialWorker::reset() {
    // Resets the serial port
    this->serialPortReset();

    // Resets the timer
    _timer.stop();

    // Resets the nonces
    _packetNonces.clear();

    // Emits a signal
    emit hasReset();
}


void WRStation::CommandSerialInterface::SerialWorker::serialPortClear() {
    // Clears the serial port and emits a signal
    bool ok = _serialPort.clear();
    if(ok) {
        emit serialPortCleared();
    }
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortClose() {
    // Closes the serial port
    _serialPort.close();
    emit serialPortClosed();
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortFlush() {
    // Flushes the serial port
    _serialPort.flush();
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortOpen(const QString& portName) {
    // Opens the port in ReadWrite mode
    _serialPort.setPortName(portName);
    bool ok = _serialPort.open(QIODevice::ReadWrite);
    if(ok)
        emit serialPortOpened();
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortReset() {
    qWarning() << "TODO: Serial Port Reset not implemented!";
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetBaudRate(qint32 baudRate) {
    // Sets the serial port baud rate
    _serialPort.setBaudRate(baudRate);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetDataBits(QSerialPort::DataBits dataBits) {
    // Sets the serial port data bits
    _serialPort.setDataBits(dataBits);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetDataTerminalReady(bool set) {
    // Sets the serial port data termnial ready
    _serialPort.setDataTerminalReady(set);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetFlowControl(QSerialPort::FlowControl flowControl) {
    // Sets the serial port flow control
    _serialPort.setFlowControl(flowControl);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetParity(QSerialPort::Parity parity) {
    // Sets the serial port parity
    _serialPort.setParity(parity);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetReadBufferSize(qint64 size) {
    // Sets the serial read buffer size
    _serialPort.setReadBufferSize(size);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetRequestToSend(bool set) {
    // Sets the serial port request to send
    _serialPort.setRequestToSend(set);
}

void WRStation::CommandSerialInterface::SerialWorker::serialPortSetStopBits(QSerialPort::StopBits stopBits) {
    // Sets the serial port stop bits
    _serialPort.setStopBits(stopBits);
}


void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortBaudRate() const {
    // Emits a signal with the baud rate
    emit serialPortBaudRateChecked(_serialPort.baudRate());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortDataBits() const {
    // Emits a signal with the data bits flag
    emit serialPortDataBitsChecked(_serialPort.dataBits());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortError() const {
    // Emits a signal with the error flag
    emit serialPortErrorChecked(_serialPort.error());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortErrorString() const {
    // Emits a signal with the error string
    emit serialPortErrorStringChecked(_serialPort.errorString());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortFlowControl() const {
    // Emits a signal with the flow control flag
    emit serialPortFlowControlChecked(_serialPort.flowControl());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortIsDataTerminalReady() {
    // Emits a signal with the data terminal ready flag
    emit serialPortIsDataTerminalReadyChecked(_serialPort.isDataTerminalReady());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortIsOpen() const {
    // Emits a signal with the request to send flag
    emit serialPortIsOpenChecked(_serialPort.isOpen());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortIsRequestToSend() {
    // Emits a signal with the request to send flag
    emit serialPortIsRequestToSendChecked(_serialPort.isRequestToSend());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortName() const {
    // Emits a signal with the port name
    emit serialPortNameChecked(_serialPort.portName());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortParity() const {
    // Emits a signal with the parity flag
    emit serialPortParityChecked(_serialPort.parity());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortReadBufferSize() const {
    // Emits a signal with the size of the buffer
    emit serialPortReadBufferSizeChecked(_serialPort.readBufferSize());
}

void WRStation::CommandSerialInterface::SerialWorker::checkSerialPortStopBits() const {
    // Emits a signal with the stop bits flag
    emit serialPortStopBitsChecked(_serialPort.stopBits());
}


void WRStation::CommandSerialInterface::SerialWorker::timerStart() {
    // Starts the timer and emits a signal
    _timer.start(_timerInterval);
    emit timerStarted();
}

void WRStation::CommandSerialInterface::SerialWorker::timerStop() {
    // Stops the timer and emits a signal
    _timer.stop();
    emit timerStopped();
}

void WRStation::CommandSerialInterface::SerialWorker::timerCheck() const {
    // Emits a signal is the timer status
    emit timerChecked(_timer.isActive());
}


void WRStation::CommandSerialInterface::SerialWorker::sendCommands() {
    if(_sendingCommands)
        return;
    _sendingCommands = true;

    // Clean old states
    _statesToSend[0].clear();

    // Requests the player states
    emit requestPlayerStates();
}

void WRStation::CommandSerialInterface::SerialWorker::addPlayerState(quint8 teamNum, quint8 playerNum, const CommandPlayerState &playerState) {
//    _statesToSend[teamNum].insert(playerNum, playerState);
    _statesToSend[0].insert(playerNum, playerState);
}

void WRStation::CommandSerialInterface::SerialWorker::playersStateCompleted() {
    // Send commands
    QHashIterator<quint8, QHash<quint8,CommandPlayerState> > teamIterator(_statesToSend);
    while(teamIterator.hasNext()) {

        // Runs the players list
        (void) teamIterator.next();
        QHashIterator<quint8, CommandPlayerState> playerIterator(teamIterator.value());
        while (playerIterator.hasNext()) {

            // Gets the team and the player numbers
            (void) playerIterator.next();
            quint8 teamNum   = teamIterator.key();
            quint8 playerNum = playerIterator.key();

            // Send player state
            QTime timer;
            timer.start();
            this->sendPlayerState(teamNum, playerNum, playerIterator.value());

            // Avoid serial overloading
            QThread::msleep(1);
        }
    }


    // Emits confirmation signal
    _sendingCommands = false;
    emit commandsSent();
}

void WRStation::CommandSerialInterface::SerialWorker::sendPlayerState(quint8 teamNum, quint8 playerNum, const WRStation::CommandPlayerState& playerState) {
    // Verifies if the serial port is open before sending the command
    if (_serialPort.isOpen()) {
        // Gets the packet nonce
        quint32 nonce = this->getNextNonce(teamNum, playerNum);

        // Assembles the packet
        QByteArray packet = _commandProtocol->encodeCommand(playerState, nonce);

        // Sends data via the serial port
        qint64 numBytes;
        numBytes = _serialPort.write(packet);
        _serialPort.waitForBytesWritten(-1);

        // Emits a signal if everything went OK
        if (numBytes == packet.size()) {
            std::cout << "[SerialWorker] Sent: player #" << (int)playerNum << " (x=" << playerState.xSpeed() << " m/s, y=" << playerState.ySpeed() << " m/s, a=" << playerState.angularSpeed() << " rad/s, kick=" << playerState.kickOnTouchEnable() << ", kickPower=" << playerState.kickOnTouchPower() << " m/s, dribbler=" << playerState.holdBallEnable() << ")\n";
            emit commandSent(teamNum, playerNum);
        } else { // Emits an error signal if not
            std::cout << "[SerialWorker] ERROR: player #" << (int)playerNum << ", " << _serialPort.errorString().toStdString()<< "\n";
            emit commandSendError(teamNum, playerNum, _serialPort.errorString());
        }
    }

    // Emits an error signal if not
    else {
        std::cout << "[SerialWorker] ERROR: player #" << (int)playerNum << ", serial device not opened!\n";
        emit commandSendError(teamNum, playerNum, QString::fromUtf8("Device not open"));
    }
}

quint32 WRStation::CommandSerialInterface::SerialWorker::getNextNonce(quint8 teamNum, quint8 playerNum) {
    // Gets the nonce
    quint32 nonce = _packetNonces[teamNum][playerNum];
    _packetNonces[teamNum][playerNum]++;
    return(nonce);
}


void WRStation::CommandSerialInterface::SerialWorker::connectSerialPortSignals() {
    // Connects serial port signals to worker signals
    QObject::connect(&_serialPort, &QSerialPort::baudRateChanged,               this, &WRStation::CommandSerialInterface::SerialWorker::serialPortBaudRateChanged);
    QObject::connect(&_serialPort, &QSerialPort::dataBitsChanged,               this, &WRStation::CommandSerialInterface::SerialWorker::serialPortDataBitsChanged);
    QObject::connect(&_serialPort, &QSerialPort::dataTerminalReadyChanged,      this, &WRStation::CommandSerialInterface::SerialWorker::serialPortDataTerminalReadyChanged);
    QObject::connect(&_serialPort, SIGNAL(error(QSerialPort::SerialPortError)), this, SIGNAL(serialPortError(QSerialPort::SerialPortError)));
    QObject::connect(&_serialPort, &QSerialPort::flowControlChanged,            this, &WRStation::CommandSerialInterface::SerialWorker::serialPortFlowControlChanged);
    QObject::connect(&_serialPort, &QSerialPort::parityChanged,                 this, &WRStation::CommandSerialInterface::SerialWorker::serialPortParityChanged);
    QObject::connect(&_serialPort, &QSerialPort::requestToSendChanged,          this, &WRStation::CommandSerialInterface::SerialWorker::serialPortRequestToSendChanged);
    QObject::connect(&_serialPort, &QSerialPort::stopBitsChanged,               this, &WRStation::CommandSerialInterface::SerialWorker::serialPortStopBitsChanged);
    QObject::connect(&_serialPort, &QSerialPort::aboutToClose,                  this, &WRStation::CommandSerialInterface::SerialWorker::serialPortAboutToClose);
    QObject::connect(&_serialPort, &QSerialPort::bytesWritten,                  this, &WRStation::CommandSerialInterface::SerialWorker::serialPortBytesWritten);
    QObject::connect(&_serialPort, &QSerialPort::readChannelFinished,           this, &WRStation::CommandSerialInterface::SerialWorker::serialPortReadChannelFinished);
    QObject::connect(&_serialPort, &QSerialPort::readyRead,                     this, &WRStation::CommandSerialInterface::SerialWorker::serialPortReadyRead);
}

void WRStation::CommandSerialInterface::SerialWorker::connectTimerSignals() {
    // Connects timer signals to worker signals
    QObject::connect(&_timer, SIGNAL(timeout()), this, SLOT(sendCommands()));
}
